#ifdef __USAGE			/* lps.c */
%C [option...][file...]
where:
-a            Enable text alignment.
-b offset     Bottom margin offset (points), default 36.
-f            Display available fonts
-h            Display the filename as a header on each page
-l            Print page in landscape mode
-o            Use left margin, default is 36 points
-p            Enable PC line drawing set
-t offset     Top margin offset (points), default 36.
-v            Verbose
-A align      Width of alignment char
-B filename   Backdrop filename
-C columns    Number of columns per physical page
-c char       Alignment char
-E            Use minimal margin defaults (1/4 inch instead of 1/2 inch)
-F font       Initial font
-I pages      Ignore this number of pages
-K copies     Number of copies to print
-L lines      Number of lines per logical page
-M pages      Logical pages per physical page(1, 2, 4, 8)
-N pages      Print this number of logical pages
-O offset     Left margin offset
-P points     Initial point size
-R offset     Right margin offset, in points
-S points     Specify extra spacing between lines, in points
-T tabsize    Specify tabsize in spaces(default 4 or TABSIZE envvar)
-X scale      Scale char width, default 1.0
-V            Virgin postscript (no prolog)
-Y scale      Scale char height, default 1.0
-Z acct_file  Update accounting file.
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>
#include <signal.h>
#include <fcntl.h>

#define	LOCAL	static


#ifndef	min
#define	min(_x,_y)	((_x) < (_y) ? (_x) : (_y))
#endif
#ifndef	max
#define	max(_x,_y)	((_x) > (_y) ? (_x) : (_y))
#endif

#define	new() if(new_flag==0) emit_backdrop(); new_flag = 1

int new_flag = 0;
extern void emit_backdrop();

	/* forward declarations */
extern void     ps_file(void);
extern int      unget(int);
extern void     break_handler(int);
extern void     save_char(int);
extern void     save_seg(int);
extern void     put_group(void);
extern void     put_text(unsigned char *, int, int);
extern void     put_header(void);
extern void     init_printer(int);
extern int      setfont(int, int, int);
extern void     handle_escape(void);
extern void     load_width_table(void);
extern void     show_page(int);
extern int      get_margin(int, double);
extern int      get(void);
extern int      get_arg(int, int, int, int);
extern void     moveto(long, int, long, long, int);
extern unsigned char *align_text(unsigned char *);
extern int      map_font(unsigned int, unsigned int);
#if 0
extern int      mappc(int *);
#else
#define mappc(x)	0
#endif

#define BASEDIR     "/usr/lib/ps/"
#define	PROLOG_FILE     BASEDIR "Prolog.ps"
#define	UPC_FILE        BASEDIR "UPC.ps"
#define	LINECHARS_FILE  BASEDIR "LineChars.ps"
#define	CHARWIDTHS_FILE BASEDIR "PSwidths.tab"


/*
 * Postscript filter program. It takes escape sequences on input and converts
 * them into postscript programs for each page. To make life easier we define
 * a few small postscript programs. ULINE  - Draws an underline of the proper
 * thickness for the current font. It does not effect the current point. ZIP
 * - An alternative show command which prints text with flair. Its kind of
 * silly. MFEED  - Requests paper from the manual feed source on the printer.
 * This is great for letter head.
 */

/*
 * Page definition constants. Note that the physical heigth and width is
 * larger than the imagable page. We apply a default margin of 1/2 inch (36
 * points).
 */
#define PT_INCH         72
#define PHY_HEIGTH	(11 * PT_INCH)           /* 11  inches * 72 points/inch */
#define PHY_WIDTH	(int)(8.5 * PT_INCH)	 /* 8.5 inches * 72 points/inch */
#define	HALF_INCH	PT_INCH/2

#define	HEADER_PSIZE	12

#define SPACE_SIZE		32000
#define MAX_FONTS		35
#define LINE_CHARS_WIDTH	1000

#define LEFT_JUSTIFIED		0
#define RIGHT_JUSTIFIED		1
#define CENTERED		2
#define FULLY_JUSTIFIED		3

/*
 * these chars must be escaped in the text, else the ps interpretor will
 * generate something funny.
 */
#define	isreserved(c) ((c) == ')' || (c) == '(' || (c) == '\\')
/*
 * Unlike many printers, postscript printers deal with families of typefaces.
 * For each family (Times-Roman) they provide the font in 4 forms. Normal,
 * Italics, Bold and Bold Italics. Going to bold is not a new font, it does
 * however require a new width table. This program is usually called by an
 * escape sequence which sets the current font to the start of a family of
 * fonts (font = 0, 4 or 8) then uses relative escape sequences to control
 * the attibutes on the font. Underline is handled differently from bold and
 * italics. The lower two bits are used as an index off the current font into
 * width tables. Underline is checked separatly.
 */

#define BOLD		(1 << 0)
#define ITALICS		(1 << 1)
#define UNDERLINE	(1 << 2)
#define MASK		(1 << 3)

/*
 * A group consists of a list of segments. Each segment may be a different
 * font or point size.  Whenever there is a change in Y (new line) then we
 * flush the segment list (group) to the page and start a new one.  In other
 * words a group consists of continuious characters with only a change in
 * attributes. A new font is considered only an attribute change. A Center or
 * justify character will also flush the group, however, it does not imply a
 * new Y co-ordinate. Note that we let the printer justify text. To do this
 * we need to provide it with a start position (via a moveto command) then a
 * series of widthshow commands. We need to know the total length of text of
 * all segments in the group before we can call widthshow. Note that the
 * group spread is kept scaled by 10,000 to prevent roundoff. Without the
 * extra power of 10 the error was noticeable on the right justified margin.
 */

struct text_seg_entry {
	struct text_seg_entry *seg_link;
	char            seg_font;	/* The font to use */
	char            seg_psize;	/* The point size to use */
	char            seg_attr;	/* Underline, Italics, Bold */
	char            seg_zing;	/* Add zing to text */
	long            seg_text_width;	/* Width in points of this segment */
	int             seg_text_nchars;	/* Number of chars in
						 * seg_text */
	int             seg_text_nspaces;	/* Number of blanks in
						 * seg_text */
	int             seg_subscript;	/* DY for sub/super script */
	char            seg_text[1];	/* Text to be printed. (,) etc have
					 * been escaped */
};



FILE           *src_in;
int				raw_flag = 0;
int				edge_flag = 0;
int				part_flag = 0;
int		src_in_virgin = 1;
int		already_postscript = 0;
char           *lp, line_buf[1000];

char            buffer[1000];
char            upc[13];
int upcinit = 0;
int upcallow = 0;

/*
 * We track the X position in 1/1000 of a point for accurate placement of a
 * char used when underlining and justification.  The Y position is not as
 * critical so we simply maintain it in points.  Note that the character
 * tables and the X position are maintained as an integer which is points *
 * 1000.  It is faster to add  integer 570  than  float .570
 */

long            x_1;		/* X position of first char of this group */
long            last_x;
int             y_1;		/* Y position of current line */
int             last_y;
int             heigth;		/* Heigth of page in points */
int             group_type;	/* Justification type for all segments */
long            group_width;	/* Width of text in all segments */
long            group_spread;	/* Amount to increase SPACE by when fully
				 * justified */
int             group_nspaces;	/* Number of spaces in the group */
int             group_not_empty;/* Does the group contain any text?  Empty
				 * lines don't */
int             gray;		/* Gray level (0 = black, 100 = white) */

/*
 * Margins in points. These margin are relative to the edge of the page they
 * apply to. The right margin (rmargin) is kept as a long in points*1000 and
 * is measured from the left of the current column.
 */

/*
 * Set margins at 1/2 inch.  These must not shrink smaller than 1/3 inch.
 */
int             tm = HALF_INCH;
int             bm = HALF_INCH;
int             lm = HALF_INCH;
long            rmargin;

/*
 * Things which can be saved and restored.
 */
struct stinfo {
	int             font;	/* Current font in use within this program */
	int             psize;	/* Current point size in use within this
				 * program */
	int             attr;	/* Three low order bits UNDERLINE | ITALICS |
				 * BOLD */
	int             zing;	/* Set non-zero for special effects */
};
LOCAL struct stinfo cur, *save_area;

LOCAL int       cur_row, cur_col;	/* Current character row and column.
					 * Used by direct cursor addressing
					 * only */

/*
 * Various flags
 */



LOCAL char      align_flag;
LOCAL char      landscape_flag;
LOCAL char      verbose_flag;
LOCAL char      dirty_flag;
LOCAL char      ink_flag;
LOCAL char      new_page;
LOCAL char      pc_lchars;
LOCAL char      offset_flag;
LOCAL char      header_flag;
LOCAL char      full_mask;
LOCAL char      show_fonts;

LOCAL int       pagemap;	/* logical to physical map */
LOCAL int       start_psize;	/* Value selected with p= option when program
				 * started */
LOCAL int       start_font;	/* Value selected with f= option when program
				 * started */
LOCAL int       last_psize;	/* Last point size selected on laser printer */
LOCAL int       last_font;	/* Last font selected on laser printer */
LOCAL int       num_columns;	/* Number of columns on the page */
LOCAL int       cur_column;	/* Current column on page */
LOCAL int       max_num_pages;	/* Number of pages to print */
LOCAL int       max_lines = 66;	/* Max number of lines per page */
LOCAL int       cur_page;	/* Current physical page which is being
				 * prepared for printing */
LOCAL int       cur_lpage;	/* Current logical  page which is being
				 * prepared for printing */
LOCAL int       num_kopies;	/* Number of copies of each page desired */
LOCAL int       ignore_num_pages;	/* Number of physical output pages to
					 * skip */
LOCAL int       xoffset = HALF_INCH;
LOCAL int       extra_spacing;
LOCAL int       mfeed_count;	/* Number of manual feed sheets */
LOCAL char      align_char = '`';
LOCAL unsigned  align_factor = 440;
LOCAL char     *fname;		/* Hold current file name being processed */
LOCAL int       tabsize = 4;	/* Default tab width */

LOCAL double    sx = 1.0, sy = 1.0;	/* X and Y scale factors for user
					 * space */

LOCAL unsigned  symbol_font;
LOCAL unsigned  dingbat_font;
LOCAL unsigned  num_fonts;
LOCAL char     *font_names[100];

LOCAL int       octx[8] = {0, PHY_WIDTH, PHY_WIDTH, PHY_WIDTH, -3 * PHY_WIDTH, PHY_WIDTH, PHY_WIDTH, PHY_WIDTH};
LOCAL int       octy[8] = {PHY_HEIGTH, 0, 0, 0, -PHY_HEIGTH + HALF_INCH, 0, 0, 0};
LOCAL int       quadx[4] = {0, PHY_WIDTH, -PHY_WIDTH, PHY_WIDTH};
LOCAL int       quady[4] = {PHY_HEIGTH - HALF_INCH, 0, -PHY_HEIGTH + HALF_INCH, 0};
LOCAL int       duplx[2] = {0, PHY_WIDTH};
LOCAL int       duply[2] = {0, 0};

LOCAL char      alloc_space[SPACE_SIZE];	/* Used to hold text_segments */
LOCAL struct text_seg_entry *alloc_ptr;	/* Ptr to current text segment be
					 * created */
LOCAL char     *text_ptr;	/* Ptr into text buffer within alloc_ptr */
LOCAL char      zing_cmd[17];	/* An alternate show command */

LOCAL char     *acct_file;
LOCAL char     *backdrop;
LOCAL char     backdrop_name[16];

#define	OPTSTR  "ab:c:fhk:lopqt:vA:B:C:EF:I:K:L:M:N:O:P:R:S:T:X:VY:Z:"
get_options(argc, argv)
	int             argc;
	char          **argv;
	{
	int             c;
	int             i;
	extern int      optind;
	extern char    *optarg;

	tm = -1;
	lm = -1;
	bm = -1;
	xoffset = -1;

	while ((c = getopt(argc, argv, OPTSTR)) != EOF) {
		switch (c) {
		case 'a':
			++align_flag;
			break;
		case 'b':
			bm = (int) strtol(optarg, NULL, 10);
			break;
		case 'c':
			if (*optarg == '\\') {
				if (optarg[1] == '\\' || optarg[1] == '\0') {
					align_char = '\\';
				} else {
					align_char = strtol(optarg + 1, NULL, 0);
				}
			} else {
				align_char = *optarg;
			}
			break;
		case 'f':
			++show_fonts;
			break;
		case 'h':
			++header_flag;
			break;
		case 'l':
			++landscape_flag;
			break;
		case 'o':
			++offset_flag;
			break;
		case 'p':
			++pc_lchars;
			break;
		case 't':
			tm = (int) strtol(optarg, NULL, 10);
			break;
		case 'v':
			++verbose_flag;
			break;

		case 'A':
			align_factor = (int) strtol(optarg, NULL, 10);
			break;
		case 'B':
			backdrop = optarg;
			break;
		case 'C':
			num_columns = (int) strtol(optarg, NULL, 10);
			break;
		case 'E':
			edge_flag = 1;
			break;
		case 'F':
			cur.font = (int) strtol(optarg, NULL, 10);
			break;
		case 'I':
			ignore_num_pages = (int) strtol(optarg, NULL, 10);
			break;
		case 'k':
			part_flag = 1;
		case 'K':
			num_kopies = (int) strtol(optarg, NULL, 10);
			break;
		case 'L':
			max_lines = (int) strtol(optarg, NULL, 10);
			break;
		case 'M':
			i = strtol(optarg, NULL, 10);
			switch (i) {
			case 2:
				pagemap = 2;
				num_columns = 2;
				sx = .647;
				sy = .773;
				break;
			case 4:
				pagemap = 4;
				num_columns = 4;
				sx = 0.5;
				sy = 0.5;
				break;
			case 8:
				pagemap = 8;
				num_columns = 8;
				sx = 0.3235;
				sy = 0.3867;
				bm /= 2;
				break;
			default:
				fprintf(stderr, "Invalid page count %d, must be 2,4,8\n", i);
				exit(EXIT_FAILURE);
			}
			break;
		case 'N':
			max_num_pages = (int) strtol(optarg, NULL, 10);
			break;
		case 'O':
			xoffset = (int) strtol(optarg, NULL, 10);
			break;
		case 'P':
			cur.psize = (int) strtol(optarg, NULL, 10);
			break;
		case 'R':
			rmargin = (int) strtol(optarg, NULL, 10) * 1000L;
			break;
		case 'S':
			extra_spacing = (int) strtol(optarg, NULL, 10);
			break;
		case 'T':
			tabsize = (int) strtol(optarg, NULL, 0);
			break;
		case 'V':
			raw_flag = 1;
			break;
		case 'X':
			sx = strtod(optarg, NULL);
			break;
		case 'Y':
			sy = strtod(optarg, NULL);
			break;
		case 'Z':
			acct_file = optarg;
			break;
		default:
			return -1;
		}
	}

	if(tm == -1) tm = edge_flag ? 18 : HALF_INCH;
	if(lm == -1) lm = edge_flag ? 18 : HALF_INCH;
	if(bm == -1) bm = edge_flag ? 18 : HALF_INCH;
	if(xoffset == -1) xoffset = edge_flag ? 18 : HALF_INCH;

	return optind;
	}


set_defaults()
{
	char           *p;
	if ((p = getenv("TABSIZE")) != NULL) {
		tabsize = strtol(p, NULL, 0);
	} else {
		tabsize = 4;
	}

	cur.font = 4;
	max_num_pages = 32000;
	num_columns = 1;
	rmargin = 468000L;
}


/*
 * These tables hold the widths of characters (at 1000 point size) for each
 * font. The table is loaded upon demand from a file.
 */

LOCAL unsigned *width_table_ptr, width_table[MAX_FONTS + 1][256];


main(argc, argv)
	char            argc, *argv[];
{
	int             lastarg;

	set_defaults();

	if (need_usage(argv))
		print_usage(argv);

	if ((lastarg = get_options(argc, argv)) < 0) {
		exit(EXIT_FAILURE);
	}
	if ((save_area = calloc(1, sizeof(struct stinfo))) == NULL) {
		fprintf(stderr, "no memory\n");
		exit(EXIT_FAILURE);
	}
	load_width_table();
	if (show_fonts) {
		int             i;
		for (i = 0; i < num_fonts; i++) {
			printf("%3d\t%s\n", i, font_names[i]);
		}
		exit(EXIT_SUCCESS);
	}
	signal(SIGINT, break_handler);

	if (cur.font >= num_fonts)
		cur.font = 0;
	width_table_ptr = width_table[cur.font];

	if (cur.psize == 0)
		cur.psize = 12;
	start_psize = cur.psize;
	start_font = cur.font;

	if (lastarg == argc) {	/* no files ... use stdout */
		src_in = stdin;
		fname = "-stdin-";
		ps_file();
	} else {
		while (lastarg < argc) {
			if ((src_in = fopen(fname = argv[lastarg++], "r")) == NULL) {
				perror(fname);
				continue;
			}
			src_in_virgin = 1;
			ps_file();
			fclose(src_in);
		}
	}
	if (acct_file) {
		int             fd;
		if ((fd = open(acct_file, O_WRONLY | O_APPEND)) != -1) {
			int             n;
			n = sprintf(line_buf, "%u\t%ld\t%s\n", cur_page, getnid(), cuserid(0));
			write(fd, line_buf, n);
		}
	}
	exit(EXIT_SUCCESS);
}

#include <ascii.h>

/*
 * Process a file
 */

void 
ps_file()
{
	int             c;

	unget(get());
	if (already_postscript) {
		while ((c=get()) != EOF) putchar(c);
		putchar('\004');
		return ;
	}
	if (landscape_flag)
		heigth = PHY_WIDTH - get_margin(tm, sy);
	else
		heigth = PHY_HEIGTH - get_margin(tm, sy);
	if(edge_flag)
		heigth -= start_psize; /* leave room for top row of chars */
	y_1 = heigth;
	if (header_flag)
		y_1 -= HEADER_PSIZE;
	cur_column = 0;
	cur.font = start_font;
	cur.psize = start_psize;
	cur.attr = cur.zing = 0;
	cur_lpage = 1;

	init_printer(1);
	alloc_ptr = (struct text_seg_entry *) alloc_space;
	text_ptr = alloc_ptr->seg_text;

	do {
		switch (c = get()) {

		case ' ':
			/* new(); */
			save_char(' ');
			++cur_col;
			break;

		case HT:
			/* new(); */
			save_char(' ');
			cur_col++;
			while (cur_col % tabsize) {
				save_char(' ');
				cur_col++;
			}
			break;

		case FF:
			/* new(); */
			save_seg(1);
			put_group();
			show_page(0);
			break;

		case LF:
			save_seg(1);
			put_group();
			cur_col = 0;
			x_1 = 0;
			y_1 -= cur.psize + extra_spacing;
			if (++cur_row < max_lines)
				break;
			show_page(0);
			break;

		case CR:
			save_seg(1);
			put_group();
			x_1 = 0;
			cur_col = 0;
			last_x = -1;
			break;
		case VT:
			save_seg(1);
			put_group();
			y_1 -= cur.psize + extra_spacing;
			if (++cur_row < max_lines)
				break;
			show_page(0);
			break;
		case ETX:	/* Center text is a ctrl c */
			save_seg(1);
			put_group();
			group_type = CENTERED;
			break;

		case ACK:	/* Fully justify is a ctrl f */
			save_seg(1);
			put_group();
			group_type = FULLY_JUSTIFIED;
			break;

		case DC2:	/* Right justify is a ctrl r */
			save_seg(1);
			last_x = -1;
			put_group();
			group_type = RIGHT_JUSTIFIED;
			break;

		case ETB:	/* Where am I */
			save_seg(1);
			last_x = -1;
			group_not_empty = 1;
			put_group();
			break;

		case EOF:
			if (alloc_ptr->seg_text_nchars) {
				save_seg(1);
				put_group();
				show_page(0);
			}
			break;

		case ESC:	/* Escape */
			handle_escape();
			break;

		case SYN:	/* SYN. On two of them we copy directly to
				 * laser printer */
			if ((c = get()) == 0x16) {
				save_seg(1);
				put_group();
				putchar('\n');
				while ((c = get()) != 0x16)
					putchar(c);
				putchar('\n');
			} else
				unget(c);
			break;

		case EOT:	/* Turn ctrl-d into a bullet */
			c = 0xb7;
		default:
			if (align_char && c == align_char) {
				++cur_col;
				while ((c = get()) == align_char)
					++cur_col;
				unget(c);
				save_seg(1);
				put_group();
				x_1 = cur_col * cur.psize;
				x_1 *= align_factor;
			} else {
				new();
				save_char(c);
				++cur_col;
			}
			break;
		}

	} while (c != EOF && max_num_pages > 0);

#if 0
	if(raw_flag == 0  ||  dirty_flag  != 0) {
		/*
		 * I think that this never makes sense unless dirty_flag != 0
		 * but I have qualified it to only make the extra check in raw
		 * mode just to be sure -gbbell-
		 */
#else
	if(ink_flag) {
#endif
		save_seg(1);
		put_group();
		show_page(1);
		}

	if(raw_flag == 0)
		printf("%%%%Trailer\n");
	putchar(0x04);		/* Send CTRL d to laser printer to end job */
	fflush(stdout);
}


void 
handle_escape()
{
	int             c, i;
	int             dx, dy;
	int             font_save, attr_save, psize_save;
	int xi, xf;
	FILE *fp;
	int n;

	/*
	 * Handle escape sequences which don't affect our X,Y position.
	 */

	save_seg(0);

	switch (c = get()) {
	case '(':
		cur.attr |= ITALICS;
		break;
	case ')':
		cur.attr &= ~ITALICS;
		break;
	case '<':
		cur.attr |= BOLD;
		break;
	case '>':
		cur.attr &= ~BOLD;
		break;
	case '[':
		cur.attr |= UNDERLINE;
		break;
	case ']':
		cur.attr &= ~UNDERLINE;
		break;
	case '{':
		cur.psize &= ~1;/* Make even */
		cur.psize = max(4, cur.psize + (start_psize / 3));
		break;
	case '}':
		cur.psize &= ~1;/* Make even */
		cur.psize = max(4, cur.psize - (start_psize / 3));
		break;
	case '#':
		cur.attr |= MASK;
		break;
	case ':':
		cur.attr &= ~MASK;
		break;
	case 'f':
		cur.font = get_arg(2, cur.font, 0, num_fonts - 1);
		break;
	case 'p':
		cur.psize = get_arg(3, cur.psize, 1, 1000);
		break;
	case 'd':
		alloc_ptr->seg_subscript -= (2 * cur.psize) / 8;
		break;
	case 'u':
		alloc_ptr->seg_subscript += (2 * cur.psize) / 8;
		break;
	case 'm':
		rmargin = get_arg(3, (int) (rmargin / 1000), 72, 999) * 1000L;
		break;
	case '`':
		save_char(c);
		++cur_col;
		break;
	case '!':
	case '@':
		get();		/* Turf colour info */
		get();
		break;
	case 'e':
		extra_spacing = get_arg(3, extra_spacing, 0, 100);
		break;
	case 'g':
		gray = get_arg(3, gray, 0, 100);
		new();
		printf("%d.%02d setgray\n", gray / 100, gray % 100);
		break;
	case 'z':		/* Add zing to the text */
		cur.zing = get_arg(1, cur.zing, 0, 4);
		if (cur.zing == 4) {
			for (i = 0; i < 16 && (c = get()) != '.'; ++i)
				zing_cmd[i] = c;
			zing_cmd[i] = '\0';
		}
		break;
	case 'k':		/* Num copies 1..9 */
		num_kopies = get_arg(1, num_kopies, 1, 9);
		break;

	case 'W':		/* Watermark/Backdrop with parts */
		part_flag = 1;
	case 'w':		/* Watermark/Backdrop */
		i = 0;
		while((c = get()) != ';'  &&  c != 0x0D  &&  c != EOF) {
			if(i < (sizeof(backdrop_name)-1))
				backdrop_name[i++] = c;
			}
		backdrop_name[i++] = 0;
		break;

	case 'o':
		xoffset = get_arg(3, xoffset, 0, PHY_HEIGTH);
		break;
	case 'a':
		align_flag = get_arg(1, align_flag, 0, 1);
		break;
	case 'F':
		full_mask = get_arg(1, full_mask, 0, 1);
		break;
	case '.':		/* Do nothing... Often used as an end marker
				 * for a number */
		break;
	case 'M':
		new();
		mfeed_count = get_arg(3, mfeed_count, 0, 999);
		printf("true mfeed\n");
		break;
	case '\n':
		unget('\r');
		break;
	case 'S':
		*save_area = cur;
		break;
	case 'R':
		cur = *save_area;
		break;
	case 'b':
		if (ignore_num_pages)
			break;
		dx = ((long) get_arg(2, 0, 0, 999) * cur.psize * align_factor) / 720 + 2;
		dy = get_arg(2, 0, 0, 999) * (cur.psize + extra_spacing) + 2;
		new();
		save_seg(1);
		put_group();
		printf("gsave\n");
		xi = x_1 / 1000;
		xf = x_1 % 1000;
		xi += get_margin(xoffset, sx);
		
		printf("%d.%.3d %d moveto ", xi - 1, xf, y_1 - 3);
		printf("%d 0 rlineto 0 %d rlineto %d 0 rlineto closepath\n", dx, dy, -dx);
		if (gray != 0)
			printf("fill\n");
		else
			printf("stroke\n");
		printf("grestore\n");
		ink_flag = 1;
		break;

	case 'U': /* UPC code */
		if (ignore_num_pages)
			break;
		memset( &upc[0], '0', 11);
		for (i = 0; i < 11 && (c = get()) != '.'; ++i)
			upc[i] = c;
		
		/*
		 * Calculate Check digit
		 */
		n = 0;
		for(i = 0; i <= 10; i += 2)
			n += (upc[i] - '0')*3;
		for(i = 1; i <= 9; i += 2)
			n += upc[i] - '0';
		upc[11] = ((10 - (n % 10)) % 10) + '0';

		new();
		save_seg(1);
		put_group();
		xi = x_1 / 1000;
		xf = x_1 % 1000;
		xi += get_margin(xoffset, sx);
		if(upcinit == 0) {
			upcinit = 1;
			if ((fp = fopen(UPC_FILE, "r")) != NULL) {
				while (n = fread(buffer, 1, sizeof(buffer), fp))
					fwrite(buffer, 1, n, stdout);
				fclose(fp);
				upcallow = 1;
				}
			}
		if(upcallow) {
			printf("%d.%.3d %d (%c) (%5.5s) (%5.5s) (%c) upcshow\n",
				xi, xf, y_1, upc[0], &upc[1], &upc[6], upc[11]);
			ink_flag = 1;
			}
		break;

	case 'c':
		c = get_arg(1, 0, 0, 2);
		save_seg(0);
		font_save = cur.font;
		attr_save = cur.attr;
		psize_save = cur.psize;
		cur.font = symbol_font;
		cur.attr = 0;
		cur.psize = (6 * cur.psize) / 8;
		width_table_ptr = width_table[cur.font];
#define	REGISTERED	0xd2
#define	COPYWRITE	0xd3
#define	TRADEMARK	0xd4
		new();
		save_char(c == 0 ? COPYWRITE : c == 1 ? REGISTERED : TRADEMARK);
		save_seg(0);
		cur.font = font_save;
		cur.attr = attr_save;
		cur.psize = psize_save;
		++cur_col;
		width_table_ptr = width_table[map_font(cur.font, cur.attr)];
		lp += 2;
		break;
	default:
		switch (c) {
		case 'A':
			--cur_row;
			break;
		case 'B':
			++cur_row;
			break;
		case 'C':
			++cur_col;
			break;
		case 'D':
			--cur_col;
			break;
		case 'H':
			cur_row = cur_col = 0;
			break;
		case '=':
		case 'Y':
			cur_row = get() - ' ';
			cur_col = get() - ' ';
			break;
		case 'y':
			save_seg(1);
			put_group();
			x_1 = get_arg(3, (int) (x_1 / 1000), 0, PHY_WIDTH) * 1000L;
			get();	/* Eat the comma */
			y_1 = get_arg(3, y_1, 0, PHY_HEIGTH);
			return;
		}

		save_seg(1);
		put_group();
		x_1 = cur_col * cur.psize;	/* Used to be start_psize
						 * here and below */
		x_1 *= align_factor;
		y_1 = heigth - (cur_row * (cur.psize + extra_spacing));
	}

	width_table_ptr = width_table[map_font(cur.font, cur.attr)];
}


/*
 * Put a group to the printer. A group is a series of congiguios characters
 * in any font with any attribute at any size. It is usually a line.
 */

void 
put_group()
	{
	register struct text_seg_entry *tp;
	long w;
	int empty;
	char *p;
	int n;

	if (new_page && ignore_num_pages <= 0 && max_num_pages > 0) {
		init_printer(0);
		new_page = 0;
		}

	if(raw_flag) {
		/*
		 * Pre-scan the entire group
		 * If ONLY spaces, then skip it entirely
		 */
		empty = 1;
		tp = (struct text_seg_entry *) alloc_space;
		for ( ; tp; tp = tp->seg_link) {
			if ((tp->seg_attr & (UNDERLINE|MASK)) == 0  &&  tp->seg_zing == 0) {
				p = tp->seg_text;
				n = tp->seg_text_nchars;
				while(n--) {
					if(*p++ != ' ')
						empty = 0;
					}
				}
			}
		if(empty)
			goto skip;
		}

	/*
	 * moveto may flush current page and setup a new page. If we have
	 * exceeded max_num_pages we just throw away this group and return.
	 * Our caller will also return and the next file (if any) will be
	 * processed.
	 */
	moveto(x_1, y_1, group_width, rmargin, group_type);

	if (max_num_pages <= 0)
		return;

	if (ignore_num_pages > 0)
		goto skip;

	tp = (struct text_seg_entry *) alloc_space;

	for ( ; tp; tp = tp->seg_link) {
		if (tp->seg_text_nchars) {
			setfont(tp->seg_font, tp->seg_attr, tp->seg_psize);
			if (tp->seg_subscript)
				printf("0 %d rmoveto ", tp->seg_subscript);
			if (tp->seg_attr & UNDERLINE) {
				w = tp->seg_text_width + (group_spread * tp->seg_text_nspaces) / 10;
				if (w > 0)
					printf("%d.%.3d %d uline ", (int) (w / 1000), (int) (w % 1000), tp->seg_psize);
				}
			if (tp->seg_attr & MASK) {
				if (full_mask)
					w = rmargin;
				else
					w = tp->seg_text_width + (group_spread * tp->seg_text_nspaces) / 10;
				if (w > 0)
					printf("%d.%.3d %d mask ", (int) (w / 1000), (int) (w % 1000), tp->seg_psize);
				}
			put_text(tp->seg_text, tp->seg_text_nchars, tp->seg_zing);
			}
		}
	
skip:
	alloc_ptr = (struct text_seg_entry *) alloc_space;
	memset(alloc_ptr, 0, sizeof(struct text_seg_entry));
	text_ptr = alloc_ptr->seg_text;
	x_1 += group_width;
	group_type = LEFT_JUSTIFIED;
	group_width = 0;
	group_spread = 0;
	group_nspaces = 0;
	group_not_empty = 0;
	}


/*
 * The only complexity here deals with text that trys to go off the end of
 * the page. We have to print the page and start a new one. This involves a
 * bit of recursion. We can't assume the user placed formfeeds in their text.
 */

void 
moveto(x, y, width, rm, type)
	long            x, width, rm;
	int             y, type;
	{
	unsigned        xi, xf;	/* Used to hold integer and fractional parts */

	if (y < bm || (max_lines && cur_row >= max_lines)) {
		show_page(0);
		if (new_page && max_num_pages > 0 && ignore_num_pages <= 0) {
			init_printer(0);
			new_page = 0;
			}
		y = y_1;
		}
	if (max_num_pages <= 0) {
		return;
		}
	if (type != LEFT_JUSTIFIED && width <= rm) {
		if (type == RIGHT_JUSTIFIED)
			x = rm - width;
		else if (type == CENTERED)
			x = (rm - width) >> 1;	/* Divide by 2 */
		else if (group_nspaces && width)
			group_spread = (10 * (rm - x - width)) / group_nspaces;
		}

	xi = x / 1000;
	xf = x % 1000;

	if ((x != last_x || y != last_y) && group_not_empty && ignore_num_pages <= 0) {
		if (cur_column == 0 ||
		    (cur_column == 2 && pagemap == 4) ||
		    (cur_column == 4 && pagemap == 8) || offset_flag)
			xi += get_margin(xoffset, sx);
		printf("%d.%.3d %d moveto ", xi, xf, y);
		last_x = x;
		last_y = y;
		}
	}


/*
 * Font will typically be 0, 4, 8 ... (start font of a set of 4). The
 * attribute is used as an index into the font names table.
 */

void 
setfont(font, attr, psize)
	int             font, attr, psize;
	{

	font = map_font(font, attr);

	if (font > num_fonts)
		font = 0;

	if (font != last_font || psize != last_psize)
		printf("/%s findfont %d scalefont setfont\n", font_names[font],
		       psize);
	last_font = font;
	last_psize = psize;
	}


/*
 * Put out one seg,ment of a group. That is, a series of characters which
 * share the same font, attributes and size.
 */

void 
put_text(text, nchars, zing)
	char           *text;
	int             nchars;
	int             zing;
	{
	char *p;
	int i;
	int ink = 0;

	if (nchars == 0)
		return;

	dirty_flag = 1;

	p = text;
	i = nchars;
	while(i--) {
		/*
		 * Set "ink" flag only if there is at least one non-space character
		 */
		if(*p++ != ' ') ink = 1;
		}

	if (group_spread) {
		if (group_spread < 0) {
			group_spread = -group_spread;
			putchar('-');
			}
		printf("%d.%.4d 0 32(", (int) (group_spread / 10000), (int) (group_spread % 10000));
		fwrite(text, 1, nchars, stdout);
		printf(")widthshow\n");
		ink_flag = 1;
		}
	else {
		if (zing) {
			putchar('(');
			fwrite(text, 1, nchars, stdout);
			if (zing == 1)
				printf(")false charpath stroke\n");
			else if (zing == 2)
				printf(")zip\n");
			else
				printf(")%s\n", zing_cmd);
			ink_flag = 1;
			}
		else if(ink) {
			putchar('(');
			fwrite(text, 1, nchars, stdout);
			printf(")show\n");
			ink_flag = 1;
			}
		}
	}

/*
 * Save away a character into the current segment.
 */


void 
save_char(c)
	int             c;
	{
	register struct text_seg_entry *tp = alloc_ptr;
	unsigned        cwidth;
	static int      font_save;

	if (c < ' ' || c >= 0x7f) {
		if (pc_lchars && cur.font != symbol_font && cur.font != dingbat_font && cur.font != num_fonts) {

			if (mappc(&c) == 0) {
				save_seg(0);
				tp = alloc_ptr;
				font_save = cur.font;
				cur.font = num_fonts;
				}
			}
		sprintf(text_ptr, "\\%3.3o", c & 0xff);
		text_ptr += 4;
		tp->seg_text_nchars += 3;	/* Plus one below */
		}
	else {
		if (pc_lchars && cur.font == num_fonts) {
			save_seg(0);
			tp = alloc_ptr;
			cur.font = font_save;
			}
		if (isreserved(c)) {
			*text_ptr++ = '\\';
			++alloc_ptr->seg_text_nchars;
			}
		if ((*text_ptr++ = c) == ' ')
			++tp->seg_text_nspaces;
		}

	++tp->seg_text_nchars;
	cwidth = width_table_ptr[c];
	tp->seg_text_width += cwidth;
	group_not_empty = 1;
	}


/*
 * Save a segment away into the current group. We don't set up the link list
 * for the last segment in a group.
 */

void 
save_seg(last_flag)
	int             last_flag;
{
	register struct text_seg_entry *tp = alloc_ptr;

	tp->seg_font = cur.font;
	tp->seg_attr = cur.attr;
	tp->seg_zing = cur.zing;
	tp->seg_psize = cur.psize;
	tp->seg_text_width *= cur.psize;
	group_width += tp->seg_text_width;
	group_nspaces += tp->seg_text_nspaces;
	alloc_ptr = (struct text_seg_entry *) text_ptr;
	tp->seg_link = last_flag ? NULL : alloc_ptr;

	tp = alloc_ptr;
	memset(tp, 0, sizeof(struct text_seg_entry));
	text_ptr = tp->seg_text;
}

/*
 * Show a page. The complexity here deals with the ability to show more than
 * one logical page on one physical page. Postscripts ability to scale,
 * translate and rotate makes this a wonderfull feature.
 */

void 
show_page(type)
	int             type;	/* 0 - normal,  1 - force to laser (set for
							 * last page) */
	{
	int             i;
	char            mfeed = 0;

	if (landscape_flag)
		heigth = PHY_WIDTH - get_margin(tm, sy);
	else
		heigth = PHY_HEIGTH - get_margin(tm, sy);
	if(edge_flag)
		heigth -= cur.psize; /* Leave room for top row of chars */
	y_1 = heigth;
	if (header_flag) {
		put_header();
		y_1 -= HEADER_PSIZE;
		}
	x_1 = 0;
	cur_row = 0;
	cur_col = 0;
	++cur_lpage;

	if (ignore_num_pages) {
		--ignore_num_pages;
		if (verbose_flag)
			fprintf(stderr, ",");
		goto skip;
		}

	--max_num_pages;
	if (verbose_flag)
		fprintf(stderr, ".");

	if (++cur_column < num_columns && type != 1) {
		switch (pagemap) {
		case 8:
			printf("%d %d translate\n", octx[cur_column], octy[cur_column]);
			break;
		case 4:
			printf("%d %d translate\n", quadx[cur_column], quady[cur_column]);
			break;
		case 2:
			printf("%d %d translate\n", duplx[cur_column], duply[cur_column]);
			break;
		default:
			if (landscape_flag) {
				printf("%d %d translate\n", (int) (PHY_HEIGTH / num_columns / sx), 0);
				}
			else if (num_columns > 2) {
				if (cur_column == 1)
					printf("%d %d translate\n", get_margin(xoffset, sx), 0);
				printf("%d %d translate\n", (int) ((PHY_WIDTH - xoffset) / num_columns / sx), 0);
				}
			else {
				printf("%d %d translate\n", (int) (PHY_WIDTH / num_columns / sx), 0);
				}
			}
		last_x = -1;
		last_y = -1;
		return;
		}

	new();

	if (dirty_flag) {
		if (mfeed_count) {
			--mfeed_count;
			++mfeed;
			/*
			 * The dumb QMS PS920 turbo clears the page on a
			 * setpapertray!! This was moved to the point where
			 * the ESC M was encountered. printf("true mfeed\n");
			 */
			}
		if (num_kopies) {
			for (i = 1; i < num_kopies; ++i) {
				if(part_flag)
					printf("(%d) partmarker ", i);
				printf("copypage\n");
				}
			if(part_flag)
				printf("(%d) partmarker ", i);
			}

		if(raw_flag)
			printf("showpage\n");
		else
			printf("showpage pg restore\n");

		if (mfeed && mfeed_count == 0)
			printf("false mfeed\n");

		new_page = 1;
		new_flag = 0;
		}
skip:
	++cur_page;
	cur_column = 0;
	dirty_flag = 0;
	ink_flag = 0;
	}


/*
 * Set the printer up. On the first call (what ADOBE calls the prologue) we
 * define our programs.
 */

void init_printer(int prologue)
	{
	int             n;
	FILE           *fp;

	if(prologue)
		printf("%%!PSAdobe-\n");

	if(raw_flag == 0  &&  prologue != 0) {

		if ((fp = fopen(PROLOG_FILE, "r")) == NULL) {
			perror(PROLOG_FILE);
			exit(-1);
			}
		while (n = fread(buffer, 1, 100, fp))
			fwrite(buffer, 1, n, stdout);
		fclose(fp);
		if (pc_lchars) {
			if ((fp = fopen(LINECHARS_FILE, "r")) == NULL) {
				perror(LINECHARS_FILE);
				exit(-1);
				}
			while (n = fread(&buffer[0], 1, sizeof(buffer), fp))
				fwrite(&buffer[0], 1, n, stdout);
			fclose(fp);
			}
		printf("%%%%EndProlog\n");
		}


	if(raw_flag == 0) {
		printf("%%%%Page: ? %d\n", cur_page);
		printf("/pg save def\n");
		}

	if (landscape_flag || pagemap == 2 || pagemap == 8)
		printf("90 rotate 0 %d translate\n", -PHY_WIDTH);

	if (sx != 1.0 || sy != sx) {
		fprintf(stdout, "%g %g scale\n", sx, sy);
	}
	switch (pagemap) {
	case 8:
		printf("%d %d translate\n", octx[0], octy[0]);
		break;
	case 4:
		printf("%d %d translate\n", quadx[0], quady[0]);
		break;
	}

	last_x = -1;
	last_y = -1;
	last_font = -1;
	last_psize = -1;
}

void emit_backdrop() {
	char *p;
	FILE *fp;
	int n;

	if(backdrop_name[0] != 0)
		p = &backdrop_name[0];
	else
		p = backdrop;

	if(p == 0  ||  *p == 0)
		return;

	sprintf(&buffer[0], "/usr/lib/ps/forms/%s.%d", p, cur_page+1 );
	fp = fopen( &buffer[0], "r");
	if(fp == NULL) {
		sprintf(&buffer[0], "/usr/lib/ps/forms/%s", p );
		fp = fopen( &buffer[0], "r");
		}

	if (fp) {
		while (n = fread(&buffer[0], 1, sizeof(buffer), fp))
			fwrite(&buffer[0], 1, n, stdout);
		fclose(fp);
		}
	}


void 
load_width_table()
{
	FILE           *fp;
	int             i;
	char            name[32];

	if ((fp = fopen(CHARWIDTHS_FILE, "r")) == NULL) {
		perror(CHARWIDTHS_FILE);
		exit(-1);
	}
	for (;;) {
		if (fread(name, 1, 32, fp) != 32)
			break;
		if ((font_names[num_fonts] = strdup(name)) == NULL) {
			fprintf(stderr, "Out of memory.\n");
			exit(EXIT_FAILURE);
		}
		if (strcmp(name, "Symbol") == 0)
			symbol_font = num_fonts;
		if (strcmp(name, "ZapfDingbats") == 0)
			dingbat_font = num_fonts;
		if (fread(&width_table[num_fonts], sizeof(int), 256, fp) != 256)
			break;
		if (++num_fonts > MAX_FONTS) {
			fprintf(stderr, "More than %d fonts in %s\n", MAX_FONTS,
				CHARWIDTHS_FILE);
			exit(EXIT_FAILURE);
		}
		font_names[num_fonts] = "LineChars";
		for (i = 0; i < 256; ++i)
			width_table[num_fonts][i] = LINE_CHARS_WIDTH;
	}
	fclose(fp);
}


map_font(font, attr)
	unsigned        font, attr;
{

	if (font < symbol_font)	/* Fonts >= symbol do not have attributes */
		return (font + (attr & 0x03));

	return (font);
}


/*
 * Called to read numbers for escape sequences.
 */

get_arg(len, cur_val, min_val, max_val)
	int             cur_val, min_val, max_val;
{
	char            relative = 0;
	int             i, c, val = 0, scale = 0;

loop:
	switch (c = get()) {
	case '-':
	case '+':
		relative = c;
		goto loop;
	case 'l':
		scale = 0;
		goto loop;
	case 'L':
		scale = 0;
		goto loop;
	}

	for (i = 0; i < len && isdigit(c); ++i) {
		val = val * 10 + (c - '0');
		c = get();
	}
	unget(c);

	if (relative == '+')
		val += cur_val;
	else if (relative == '-')
		val = cur_val - val;

	if (scale)
		val = val * scale;

	return (max(min(max_val, val), min_val));
}


/*
 * This routine takes a buffer to a line and tries to add alignment
 * characters at reasonable places.  This ensures that tables come out
 * correctly when using a proportional font.
 */

char           *
align_text(buffer)
	char           *buffer;
{
	register char  *p = buffer, *p1;
	char            dot;

	if ((*p < ' ' && *p != 0x1b) || cur.font == 0)	/* Don't align justified
							 * lines etc... */
		return (buffer);

	/*
	 * Process leading blanks. We step over points which start with a
	 * dash (-) or a bullet (o).
	 */

	if (*p == '-' || *p == 'o')
		++p;
	if (*p == ' ') {
		while (*p == ' ')
			++p;

		*(p + -1) = align_char;
	}
	/*
	 * Process runs of blanks (except  ".  ") Process " - "  and  " o "
	 * in case there are multiple columns of points. Process  ddd)  where
	 * d is a digit.
	 */

	while (*p) {
		if (*p == ' ') {
			if (*(p + 1) == ' ' && dot == 0) {
				while (*p == ' ')
					++p;
				*(p + -1) = align_char;
			} else if ((*(p + 1) == '-' || *(p + 1) == 'o') && *(p + 2) == ' ') {
				*p++ = align_char;
				*++p = align_char;
				dot = 0;
			} else if (dot)
				--dot;
		} else if (*p == '.' || *p == '!' || *p == '?')
			dot = 1;
		else if (*p == ')' && *(p + 1) == ' ') {
			p1 = p - 1;
			while (p1 >= buffer && *p1 >= '0' && *p1 <= '9')
				--p1;
			if (p1 < buffer || *p1 == ' ' || *p1 == align_char)
				*++p = align_char;
		} else if (*p == 0x04)
			return (buffer);
		else
			dot = 0;

		++p;
	}

	return (buffer);
}



get()
{
	if (lp == NULL || *lp == '\0') {
		if (fgets(lp = line_buf, 1000, src_in) == 0)
			return (EOF);
		if (src_in_virgin) {
			already_postscript = (*lp == '%' && lp[1] == '!');
			src_in_virgin = 0;
		}
		if (align_flag)
			align_text(line_buf);
	}
	return (*lp++ & 0xff);
}


void 
unget(c)
	int             c;
{

	if (c != EOF)
		*--lp = c;
}


/*
 * This is still very primitive. It will be used to put out file names,
 * dates, page numbers etc... on the top of each page. It will of course be
 * optional since output will often be from word processors.
 */

void 
put_header()
{
	register char  *s;

	s = "gsave\n\
		 /%s findfont 12 scalefont setfont\n\
		 72 %d moveto (%s  Page %d)show\n\
		 grestore\n";

	printf(s, font_names[start_font], heigth, fname, cur_lpage);
}


void 
break_handler(int sig_no)
{
	int i;
	i = sig_no;		/* shut compiler up */

	putchar(0x03);
	putchar(0x04);
	putchar(0x04);
	fflush(stdout);
	exit(-1);
}

get_margin(points, scale)
	int             points;
	double          scale;
{

	/*
	 * If -o flag is set (indicating left margin always used)
	 * then allow ANY value for x offset (-O xxx)
	 */
	if(offset_flag  ||  edge_flag)
		return(points);

	if (landscape_flag) {
		if (points * scale < 72 / 4)
			return ((72 / 4) / scale);
	} else {
		if (points * scale < 72 / 3)
			return ((72 / 3) / scale);
	}

	return (points);
}

#if 0


char            pctab[] = {
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
	'', '?',
};

mappc(cp)
	int            *cp;
{
	int             c = *cp;
	register char  *p;

	for (p = pctab; *p; p += 2)
		if (*p == c) {
			*cp = *(p + 1);
			return (1);
		}
	return 0;
}
#endif
